import logging
from typing import Optional
from linebot import LineBotApi
from linebot.models import TextSendMessage


def setup_logger(log_file_name:Optional[str]=None,line_channel_access_token:Optional[str]=None):
    l = logging.getLogger()
    
    if l.hasHandlers(): return

    formatter = logging.Formatter('%(asctime)s| %(message)s', datefmt='%d/%b/%Y-%H:%M:%S')
    streamHandler = logging.StreamHandler()
    streamHandler.setFormatter(formatter)
    l.addHandler(streamHandler)  

    if log_file_name is not None:
        fileHandler = logging.FileHandler(log_file_name, mode='a')
        fileHandler.setFormatter(formatter)
        l.addHandler(fileHandler)
    if line_channel_access_token is not None:
        lineHandler = LineBotHandler(line_channel_access_token)
        # lineHandler.setFormatter(formatter)
        l.addHandler(lineHandler)
        
    l.setLevel("INFO")


def log(msg:str):
    l = logging.getLogger()
    l.info(msg)


class LineBotHandler(logging.Handler):
    def __init__(self,channel_access_token):
        logging.Handler.__init__(self)
        self.line_bot = LineBotApi(channel_access_token)

    def emit(self,record):
        msg = self.format(record)
        self.line_bot.broadcast(TextSendMessage(msg))


from functools import reduce
import os


def linux_beep(duration:float=0.1,frequency:int=300):
    os.system(f"play -nq -t alsa synth {duration} sine {frequency}")

def compose(*func):
    def compose_function(f, g):
        return lambda *args : g(f(*args))
    return reduce(compose_function, func)


from petch.logging import log
import cProfile
import pstats
from time import time

def timer_func(func):
    def wrap_func(*args, **kwargs):
        t1 = time()
        result = func(*args, **kwargs)
        t2 = time()
        print(f"Function {func.__name__!r} executed in {(t2-t1):.4f}s")
        return result
    return wrap_func


def profile_function(func):
    def wrapper(*args, **kwargs):
        with cProfile.Profile() as pr:
            result = func(*args, **kwargs)
            stats = pstats.Stats(pr)
        stats.sort_stats(pstats.SortKey.TIME)
        stats.dump_stats(filename="profiling.prof")
        return result
    return wrapper

def log_funtion_and_input_output(func):
    def create_text(inputs):
        try: return '\n'.join([str(input) for input in inputs])
        except TypeError: return f"{str(inputs)}"
    def wrap(*args,**kwargs):
        results = func(*args,**kwargs)
        log(f"{create_text([func.__name__,'Input:',create_text(args),create_text(kwargs),'Output:',create_text(results)])}")
        return results
    return wrap


from typing import Any, Callable
from petch.betterlist import BetterList


class BetterDict(dict):
    '''
    This Dict Class allows you todo javascript styled continuous operation.
    ex.
    x = BetterDict({'a':1,'b':2,'c':3,'d':4})
    y = x.filter_key(func1).filter_value(func2).map_key(func3).foreach_value(print)
    '''
    @property
    def keys_list(self):
        return BetterList(self.keys())

    @property
    def values_list(self):
        return BetterList(self.values())

    @property
    def length(self):
        return len(self)

    def sort(self,func:Callable=lambda x:x[1]):
        return BetterDict(sorted(self.items(),key=func))

    def max(self,n_largest:int=1,func:Callable=lambda x:x[1]):
        """
        BetterDict({"a":1,"b":2,"c":3,"d":4})
        Return key list that looks like this
        ['d', 'c']
        """
        return BetterList(self.sort(func).keys_list[-n_largest:][::-1])

    def min(self,n_smallest:int=1,func:Callable=lambda x:x[1]):
        """
        BetterDict({"a":1,"b":2,"c":3,"d":4})
        Return key list that looks like this
        ['a', 'b']
        """
        return BetterList(self.sort(func).keys_list[:n_smallest])

    def foreach(self,func:Callable):
        for key,value in self.items():
            func(key,value)

    def foreach_key(self,func:Callable):
        for key,_ in self.items():
            func(key)

    def foreach_value(self,func:Callable):
        for _,value in self.items():
            func(value)

    def filter_key(self,func:Callable):
        return BetterDict({key:value for key,value in self.items() if func(key)})

    def filter_out_key(self,func:Callable):
        return BetterDict({key:value for key,value in self.items() if not func(key)})

    def filter_value(self,func:Callable):
        return BetterDict({key:value for key,value in self.items() if func(value)})

    def filter_out_value(self,func:Callable):
        return BetterDict({key:value for key,value in self.items() if not func(value)})

    def map(self,func:Callable):
        return BetterDict({key:func(key, value) for key,value in self.items()})

    def map_key(self,func:Callable):
        return BetterDict({func(key):value for key,value in self.items()})

    def map_value(self,func:Callable):
        return BetterDict({key:func(value) for key,value in self.items()})

    def reduce_key(self,func:Callable):
        return self.keys_list.reduce(func)

    def reduce_value(self,func:Callable):
        return self.values_list.reduce(func)



class Event:
    def __init__(self):
        self.listeners = []

    def __iadd__(self, listener):
        """Shortcut for using += to add a listener."""
        self.listeners.append(listener)
        return self

    def __isub__(self, listener):
        """Shortcut for using -= to add a listener."""
        self.listeners.remove(listener)
        return self

    def notify(self, *args, **kwargs):
        for listener in self.listeners:
            listener(*args, **kwargs)
            
    async def async_notify(self, *args, **kwargs):
        for listener in self.listeners:
            await listener(*args, **kwargs)



from typing import Any, Callable
from functools import reduce

class BetterList(list):
    '''
    This List Class allows you to do javascript styled continuous operation.
    ex.
    x = BetterList([1,2,3,4])
    y = x.filter(func1).map(func2).reduce(func3)
    z = x.map(func4).for_each(func5)
    '''
    @property
    def length(self):
        return len(self)

    def sort(self,func:Callable=lambda x:x):
        return BetterList(sorted(self,key=func))

    def min(self,n_smallest:int=1,func:Callable=lambda x:x):
        """
        Return looks like this:
        the min one is the first item
        [0, 0, 4, 5, 5, 5, 6, 10, 11, 12]
        """
        return BetterList(self.sort(func)[:n_smallest])

    def max(self,n_largest:int=1,func:Callable=lambda x:x):
        """
        Return looks like this:
        the max one is the first item
        [100, 99, 98, 97, 96, 95, 94, 93, 93, 92]
        """
        return BetterList(self.sort(func)[-n_largest:][::-1])
    
    def filter(self,func:Callable):
        return BetterList(filter(func,self))

    def filter_out(self,func:Callable):
        return BetterList(filter(lambda x: not(func(x)),self))

    def remove_duplicate(self):
        return BetterList(set(self))

    def map(self,func:Callable):
        return BetterList(map(func,self))
    
    def reduce(self,func:Callable):
        return reduce(func,self)

    def foreach(self,func:Callable):
        for x in self:
            func(x)

    def filter_index(self,func:Callable):
        _index = BetterList([i for i,_ in enumerate(self)])
        return BetterList([self[i] for i in _index.filter(func)])

    def get(self,index,default_value=None):
        try: return self[index]
        except IndexError: return default_value

    def value_after(self,value:Any):
        it = iter(self)
        x = ""
        while str(x) != str(value) or str(x) not in str(value):
            x=next(it)
        return next(it)

    def first(self,n:int):
        return BetterList(range(0,n)).map(lambda i: self[i])

    def last(self,n:int):
        return BetterList(range(-n,0)).map(lambda i: self[i])
